#!/usr/bin/env python3
# coding=utf-8
'''
 * @Author: z417
 * @Date: 2020-11-17 03:44:42
 * @LastEditors: z417
 * @LastEditTime: 2020-11-17 09:51:38
 * @FilePath: /home/vscode/Yuri/sql/testORM.py
 * @Description: file desc
'''
import aiomysql


class Dict(dict):
    '''dict子类, 扩展了value的访问方式; 还支持传入两个长度相等的tuple, 组成key-value'''

    def __init__(self, names=(), values=(), **kw):
        '''
        @names: tuple形式的key集合
        @values: tuple形式的value集合
        '''
        super().__init__(**kw)
        for k, v in zip(names, values):
            self[k] = v

    def __getattr__(self, item):
        '''
        当试图访问实例不存在的属性时, 会自动调用该方法; 访问方式就是'点'
        '''
        try:
            return self[item]
        except KeyError:
            raise AttributeError(r"'Dict' object has no attribute '%s'" % item)

    def __setattr__(self, key, value):
        '''
        当试图给不存在的属性赋值时, 会自动调用该方法
        '''
        self[key] = value


def createArgsStr(num):
    return ', '.join(['?' for _ in range(num)])


class ModelMetaClass(type):
    def __new__(cls, name, bases, attrs):
        if name == 'Model':
            # 当出现与'Model'同名的类时, 直接创建这类
            return type.__new__(cls, name, bases, attrs)

        # 定义表名: 要么在类中定义__table__属性(目的是"类名可以与表名不相同"), 否则与类名相同
        tableName = attrs.get('__table__') or name
        print(f'建立映射关系: {name}类 --> {tableName}表')

        mappings = Dict()   # 存储column与Field 子类的对应关系, Field在上一章中定义的, 忘了回去翻
        fields = []         # 用来存储除主键以外的所有字段名
        primaryKey = None   # 用来记录主键字段的名字, 初始没有

        for k, v in attrs.items():
            # 遍历所有属性, 即映射表的字段, 读不懂请回看第二章 Users 的定义
            if isinstance(v, Field):     # Field类, 所有字段类型的父类
                print(f'建立映射... column: {k} ==> class: {v}')
                mappings[k] = v

                if v.primaryKey:         # 判断字段是否被设置成了主键
                    if primaryKey:       # 因为一张表只能有一个主键
                        raise Exception(f'Duplicate primary key for field {k}')
                    primaryKey = k
                else:
                    fields.append(k)

        if not primaryKey:               # 这里做了一步强制要求设置主键, 你也可以去掉
            raise Exception(f'请给表{tableName}设置主键')

        for k in mappings.keys():
            # 删除原属性, 避免实例的属性遮盖类的同名属性, 况且我们已经保存到 mappings 中了, 不怕丢
            attrs.pop(k)

        # 接下来给本元类(ModelMetaClass)创建的class(如 Model)设置私有属性
        attrs['__mappings__'] = mappings
        attrs['__table__'] = tableName
        attrs['__primaryKey__'] = primaryKey
        attrs['__fields__'] = fields

        # 以下 8 行是新增的内容(注意提前定义好 createArgsStr), update 单拎出来, 这里不做处理
        escapedFields = list(map(lambda a: f'`{a}`', fields))
        attrs['__select__'] = f"SELECT `{primaryKey}`, {', '.join(escapedFields)} FROM `{tableName}`"
        attrs['__delete__'] = f"DELETE FROM `{tableName}` WHERE `{primaryKey}`=?"

        # 由于 update_at & created_at 可以由mysql自动写入, 我们写入时可以不传
        fields.remove('updated_at')
        fields.remove('created_at')
        attrs['__fields_2__'] = fields
        escapedFields_2 = list(map(lambda a: f'`{a}`', fields))
        attrs['__insert__'] = f"INSERT INTO `{tableName}` ({', '.join(escapedFields_2)}) VALUES ({createArgsStr(len(escapedFields_2))})"

        return type.__new__(cls, name, bases, attrs)


class Model(Dict, metaclass=ModelMetaClass):
    """指定metaclass, 以实现动态定制"""

    def __init__(self, **kw):
        super().__init__(**kw)

    def getValue(self, key):
        return getattr(self, key, None)

    def getValueOrDefault(self, key):
        value = getattr(self, key, None)
        if value is None:
            field = self.__mappings__[key]   # 从所有column中获取value
            if field.default is not None:
                # 如果default指向是方法(如time.time), 则调用方法获取其值; 否则直接赋值
                value = field.default() if callable(field.default) else field.default
                print(f'using defalut value for {key}: {value}')
                setattr(self, key, value)    # 其实是调 Dict.__setattr__, 以支持用"."访问
        return value

    # 以下内容为更新(新增)
    @classmethod
    def appendCondition(cls, sql, where, args, **kw):
        if where:
            sql.append('WHERE')
            sql.append(where)
        if not args:
            args = []
        orderBy = kw.get('orderBy')
        if orderBy:
            sql.append('ORDER BY')
            sql.append(orderBy)
        limit = kw.get('limit')
        if limit:
            sql.append('LIMIT')
            if isinstance(limit, int):
                sql.append('?')
                args.append(limit)
            elif isinstance(limit, tuple) and len(limit) == 2:
                sql.append('?, ?')
                args.extend(limit)  # 在列表末尾一次性追加另一个序列中的多个值
            else:
                raise Exception(f'Invalid limit value: {limit}')
        return sql

    @classmethod
    async def findAll(cls, where=None, args=None, **kw):
        '''返回结果集'''
        sql = [cls.__select__]
        sql = cls.appendCondition(sql, where, args, **kw)
        rs = await Mysql.select(' '.join(sql), args)
        # cls(**r)是调用Model.__init__方法, 将dict转为Dict, 就可以 . 的方式获取value
        return [cls(**r) for r in rs]

    @classmethod
    async def findFieldValue(cls, selectField, where=None, args=None, **kw):
        sql = [f"SELECT {selectField} FROM `{cls.__table__}`"]
        sql = cls.appendCondition(sql, where, args, **kw)
        rs = await Mysql.select(' '.join(sql), args)
        return [cls(**r) for r in rs]

    @classmethod
    async def findByPrimaryKey(cls, pk):
        '''find object by primary key.'''
        rs = await Mysql.select(f'{cls.__select__} where `{cls.__primaryKey__}`=?', [pk], 1)
        if len(rs) == 0:
            return {}
        return cls(**rs[0])

    async def insert(self):
        # 因为insert都是具体的一行数据, 即单个实例, 所以没有定义成类方法
        args = list(map(self.getValueOrDefault, self.__fields_2__))
        rows = await Mysql.execute(self.__insert__, args)
        if rows != 1:
            print(f'插入数据失败, 影响行数: {rows}')

    @classmethod
    async def update(cls, setField, where, args):
        sql = f"UPDATE `{cls.__table__}` SET {', '.join(map(lambda a: f'`{a}`=?', setField))} WHERE {where}"
        rows = await Mysql.execute(sql, args)
        if not rows:
            print(f'failed to update by {where}, affected rows: {rows}')

    @classmethod
    async def delete(cls, pk):
        """delete by primaryKey"""
        rows = await Mysql.execute(cls.__delete__, [pk])
        if rows != 1:
            print(
                f'failed to delete by {cls.__primaryKey__}, affected rows: {rows}')

    @classmethod
    async def logicDelete(cls, pk):
        await cls.update(setField=['is_deleted'],
                         where=f'{cls.__primaryKey__}=?', args=[1, pk])


class Field:
    """映射字段名、字段类型和默认值"""

    def __init__(self, columnType, primaryKey, default):
        self.columnType = columnType
        self.primaryKey = primaryKey
        self.default = default

    def __str__(self):
        """
        该方法在Field的实例被print时调用,这样自定义后,就不会打印内存地址了,
        而是打印‘<类名: columnType>’, 这个特性在后面打印信息时才会用到, 当前看不懂请先忽略
        """
        # 这是python3.7及以后新增的写法，等于'<{}: {}>'.format(self.__class__.__name__, self.columnType)
        return f'<{self.__class__.__name__}: {self.columnType}>'


class StrField(Field):
    """映射字符类型,如char和varchar"""

    def __init__(self, primaryKey=False, default=None, ddl='varchar(100)'):
        """
        主键、默认值和类型三要素,在实例化时候可传入参数定制,这里先给了默认值
        """
        super().__init__(ddl, primaryKey, default)


class BooleanField(Field):
    """映射boolean类型"""

    def __init__(self, default=False):
        """
        mysql布尔类型关键字:boolean
        布尔类型不可能做主键，所以primaryKey写死False
        默认值False
        """
        super().__init__('boolean', False, default)


class IntegerField(Field):
    """映射整型,如int"""

    def __init__(self, primaryKey=False, default=0, ddl='bigint'):
        """实例化时可通过参数定制3要素"""
        super().__init__(ddl, primaryKey, default)


class FloatField(Field):
    """映射浮点类型,如float, real"""

    def __init__(self, primaryKey=False, default=0.0, ddl='real'):
        super().__init__(ddl, primaryKey, default)


class TextField(Field):
    """映射大文本类型,如text"""

    def __init__(self, default=None, ddl='text'):
        """一般不会做主键,所以primaryKey写死False"""
        super().__init__(ddl, False, default)


class DateTimeField(Field):
    """映射时间类型,如timestamp, date, datetime"""

    def __init__(self, default=None, ddl='timestamp'):
        """一般不会做主键,所以primaryKey写死False"""
        super().__init__(ddl, False, default)


def toDict(d: dict):
    D = Dict()
    for k, v in d.items():
        D[k] = toDict(v) if isinstance(v, dict) else v
    return D


class Mysql:
    @staticmethod
    async def createPool():
        '''连接池由全局变量__pool存储，缺省情况下将编码设置为utf8，自动提交事务'''
        try:
            global __pool
            __pool = await aiomysql.create_pool(
                host='localhost',        # mysql服务器地址
                port=3306,               # mysql监听端口
                user='awesome',          # mysql用户名
                password='123456',       # mysql密码
                db='awesome',            # mysql实例名(database)
                charset='utf8',          # 编码设置
                autocommit=True,         # 自动提交insert/update/delete
                maxsize=10,              # 连接池上限
                minsize=1                # 连接池下限
            )
            print(f'创建成功{type(__pool)}')
        except Exception as e:
            print(f'连接 mysql 出错:{e}')

    @staticmethod
    async def select(sql, args, size=None):
        '''传入SQL语句和SQL参数
        @sql :  sql语句
        @args:  条件值列表
        @size:  查询的结果集大小
        '''
        print(f'{sql} ==> {args}')
        # 使用with上下文管理器, 自动执行close
        async with __pool.acquire() as conn:
            # acquire: 从空闲池获得连接。如果需要，并且池的大小小于maxsize，则创建新连接。
            async with conn.cursor(aiomysql.DictCursor) as cur:
                # cursor: 连接的游标
                # DictCursor: 作为字典返回结果的游标
                await cur.execute(sql.replace('?', '%s'), args or ())
                # sql.replace('?', '%s'): 'SELECT * FROM users WHERE uid=?'.replace('?', '%s')
                # args or (): 'SELECT * FROM users WHERE uid=%s' % (101,)
                if size:
                    rs = await cur.fetchmany(size)
                else:
                    rs = await cur.fetchall()
            print(f'rows returned: {len(rs)}')
            return rs

    @staticmethod
    async def execute(sql, args, autocommit=True):
        '''
        INSERT、UPDATE、DELETE操作，可以通用的execute函数，
        因为这3种SQL的执行都需要相同的参数, 返回一个整数表示影响的行数
        '''
        print(f'{sql} ==> {args}')
        async with __pool.acquire() as conn:
            if not autocommit:
                await conn.begin()           # 开始处理
            try:
                async with conn.cursor(aiomysql.DictCursor) as cur:
                    await cur.execute(sql.replace('?', '%s'), args)
                    affected = cur.rowcount  # 受影响的行数
                if not autocommit:
                    await conn.commit()      # 提交更改, 非自动提交时才需要
            except Exception as e:
                if not autocommit:
                    await conn.rollback()    # 回滚更改
                raise Exception(e)
            return affected


class users(Model):
    """
    继承自Model, 这样Users就有了Dict特性, 同时在实例化Users时, 又会以ModelMetaClass定制的特性创建
    """
    uid = IntegerField(primaryKey=True, ddl='int(11)')
    email = StrField(ddl='varchar(50)')
    passwd = StrField(ddl='char(32)')
    admin = IntegerField(default=0, ddl='tinyint(1)')
    name = StrField(ddl='varchar(50)')
    birthday = DateTimeField(ddl='DATE')
    user_status = IntegerField(default=0, ddl='tinyint(1)')
    image = StrField(default='about:blank', ddl='varchar(500)')
    created_at = DateTimeField(ddl='timestamp')
    updated_at = DateTimeField(ddl='timestamp')
    created_by = IntegerField(ddl='int(11)')
    updated_by = IntegerField(ddl='int(11)')
    is_deleted = IntegerField(default=0, ddl='tinyint(1)')


if __name__ == '__main__':
    import asyncio
    loop = asyncio.get_event_loop()
    loop.run_until_complete(Mysql.createPool())
    u = users()
    rs = loop.run_until_complete(u.findAll())
    for i in rs:
        print(i.uid)

    rs = loop.run_until_complete(u.findAll(where='is_deleted=?', args=[1]))
    print(rs[0].name)

    rs = loop.run_until_complete(
        u.findAll(where='is_deleted=?', args=[1], orderBy='uid DESC'))
    print(rs)

    rs = loop.run_until_complete(u.findAll(where='is_deleted=?', args=[
                                 1], orderBy='uid DESC', limit=1))
    print(rs[0].email)

    rs = loop.run_until_complete(u.findFieldValue(
        'uid, name, email', 'is_deleted=?', [1], orderBy='uid DESC', limit=(0, 2)))
    print(rs)

    rs = loop.run_until_complete(u.findByPrimaryKey(102))
    print(rs)

    import hashlib
    passwd = hashlib.md5('123456'.encode(encoding='utf-8')).hexdigest()
    test = users(email='test@ztest.top', passwd=passwd, name='test',
                 created_by=100, updated_by=100)
    loop.run_until_complete(test.insert())

    loop.run_until_complete(
        u.update(setField=['name'], where='uid=106', args=['newTest']))
    loop.run_until_complete(u.logicDelete(106))
    print(loop.run_until_complete(u.findByPrimaryKey(106)))
